from tfrddlsim.rddl2tf.fluentshape import TensorFluentShape

import unittest


class TestTensorFluentShape(unittest.TestCase):

    def test_broadcast(self):
        tests = [
            (TensorFluentShape([], False), TensorFluentShape([], False), None, None),
            (TensorFluentShape([8], False), TensorFluentShape([], False), None, None),
            (TensorFluentShape([], False), TensorFluentShape([8], False), None, None),
            (TensorFluentShape([8, 8], False), TensorFluentShape([8], False), None, None),
            (TensorFluentShape([8], False), TensorFluentShape([8, 8], False), None, None),
            (TensorFluentShape([100], True), TensorFluentShape([100], True), None, None),
            (TensorFluentShape([100, 8], True), TensorFluentShape([100], True), None, [100, 1]),
            (TensorFluentShape([100], True), TensorFluentShape([100, 8], True), [100, 1], None),
            (TensorFluentShape([100, 8, 8], True), TensorFluentShape([100], True), None, [100, 1, 1]),
            (TensorFluentShape([100], True), TensorFluentShape([100, 8, 8], True), [100, 1, 1], None),
            (TensorFluentShape([100, 8, 8], True), TensorFluentShape([100, 8], True), None, [100, 1, 8]),
            (TensorFluentShape([100, 8], True), TensorFluentShape([100, 8, 8], True), [100, 1, 8], None),
            (TensorFluentShape([100], True), TensorFluentShape([], False), None, None),
            (TensorFluentShape([], False), TensorFluentShape([], True), None, None),
            (TensorFluentShape([100], True), TensorFluentShape([], False), None, None),
            (TensorFluentShape([100], True), TensorFluentShape([8], False), [100, 1], None),
            (TensorFluentShape([8], False), TensorFluentShape([100], True), None, [100, 1]),
            (TensorFluentShape([100], True), TensorFluentShape([8, 7], False), [100, 1, 1], None),
            (TensorFluentShape([8, 7], False), TensorFluentShape([100], True), None, [100, 1, 1]),
            (TensorFluentShape([100, 8], True), TensorFluentShape([], False), None, None),
            (TensorFluentShape([], False), TensorFluentShape([100, 8], True), None, None),
            (TensorFluentShape([100, 8], True), TensorFluentShape([8], False), None, None),
            (TensorFluentShape([8], False), TensorFluentShape([100, 8], True), None, None),
            (TensorFluentShape([100, 8, 7], True), TensorFluentShape([7], False), None, [1, 7]),
            (TensorFluentShape([7], False), TensorFluentShape([100, 8, 7], True), [1, 7], None),
            (TensorFluentShape([100, 7, 8], True), TensorFluentShape([7, 8], False), None, None),
            (TensorFluentShape([7, 8], False), TensorFluentShape([100, 7, 8], True), None, None),
            (TensorFluentShape([8, 8], False), TensorFluentShape([100, 8], True), None, [100, 1, 8]),
            (TensorFluentShape([100, 8], True), TensorFluentShape([8, 8], False), [100, 1, 8], None),
            (TensorFluentShape([2, 2], False), TensorFluentShape([1, 2], True), None, [1, 1, 2]),
            (TensorFluentShape([1, 2], True), TensorFluentShape([2, 2], False), [1, 1, 2], None),
        ]

        for s1, s2, ss1, ss2 in tests:
            reshape1, reshape2 = TensorFluentShape.broadcast(s1, s2)
            if ss1 is None:
                self.assertIsNone(reshape1)
            else:
                self.assertListEqual(reshape1, ss1)

            if ss2 is None:
                self.assertIsNone(reshape2)
            else:
                self.assertListEqual(reshape2, ss2)



from pyrddl.rddl import RDDL
from pyrddl.parser import RDDLParser
from pyrddl.expr import Expression
from pyrddl import utils

from tfrddlsim.rddl2tf.compiler import Compiler
from tfrddlsim.rddl2tf.fluent import TensorFluent
from tfrddlsim.rddl2tf.fluentshape import TensorFluentShape

import numpy as np
import tensorflow as tf

import unittest


class TestCompiler(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        with open('rddl/Reservoir.rddl', mode='r') as file:
            RESERVOIR = file.read()

        with open('rddl/Mars_Rover.rddl', mode='r') as file:
            MARS_ROVER = file.read()

        parser = RDDLParser()
        parser.build()

        cls.rddl1 = parser.parse(RESERVOIR)
        cls.rddl2 = parser.parse(MARS_ROVER)

    def setUp(self):
        self.compiler1 = Compiler(self.rddl1)
        self.assertIs(self.compiler1.rddl, self.rddl1)
        self.compiler2 = Compiler(self.rddl2)
        self.assertIs(self.compiler2.rddl, self.rddl2)

    def test_build_object_table(self):
        self.assertIn('res', self.compiler1.object_table)
        size = self.compiler1.object_table['res']['size']
        idx = self.compiler1.object_table['res']['idx']
        self.assertEqual(size, 8)
        objs = ['t1', 't2', 't3', 't4', 't5', 't6', 't7', 't8']
        for i, obj in enumerate(objs):
            self.assertIn(obj, idx)
            self.assertEqual(idx[obj], i)

    def test_build_action_preconditions_table(self):
        local_preconds = self.compiler1.local_action_preconditions
        self.assertIsInstance(local_preconds, dict)
        self.assertEqual(len(local_preconds), 1)
        self.assertIn('outflow/1', local_preconds)
        self.assertEqual(len(local_preconds['outflow/1']), 2)

        global_preconds = self.compiler1.global_action_preconditions
        self.assertIsInstance(global_preconds, list)
        self.assertEqual(len(global_preconds), 0)

    def test_lower_bound_constraints(self):
        lower_bounds = self.compiler1.action_lower_bound_constraints
        self.assertIsInstance(lower_bounds, dict)
        self.assertIn('outflow/1', lower_bounds)
        lower = lower_bounds['outflow/1']
        self.assertIsInstance(lower, Expression)
        self.assertTrue(lower.is_number_expression())
        self.assertEqual(lower.value, 0)

    def test_upper_bound_constraints(self):
        upper_bounds = self.compiler1.action_upper_bound_constraints
        self.assertIsInstance(upper_bounds, dict)
        self.assertIn('outflow/1', upper_bounds)
        upper = upper_bounds['outflow/1']
        self.assertIsInstance(upper, Expression)
        self.assertTrue(upper.is_pvariable_expression())
        self.assertEqual(upper.name, 'rlevel/1')

    def test_compile_action_preconditions(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler2]
        expected_preconds = [2, 1]
        for compiler, expected in zip(compilers, expected_preconds):
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            preconds = compiler.compile_action_preconditions(state, action)
            self.assertIsInstance(preconds, list)
            self.assertEqual(len(preconds), expected)
            for p in preconds:
                self.assertIsInstance(p, TensorFluent)
                self.assertEqual(p.dtype, tf.bool)
                self.assertEqual(p.shape.batch_size, batch_size)
                self.assertTrue(p.shape.batch)
                self.assertTupleEqual(p.shape.fluent_shape, (1,))

    def test_compile_action_preconditions_checking(self):
        batch_size = 1000
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            compiler.batch_mode_on()
            state = compiler.compile_initial_state(batch_size)
            action = compiler.compile_default_action(batch_size)
            checking = compiler.compile_action_preconditions_checking(state, action)
            self.assertIsInstance(checking, tf.Tensor)
            self.assertEqual(checking.dtype, tf.bool)
            self.assertListEqual(checking.shape.as_list(), [batch_size])

    def test_compile_action_bound_constraints(self):
        batch_size = 1000
        self.compiler1.batch_mode_on()
        initial_state = self.compiler1.compile_initial_state(batch_size)
        default_action_fluents = self.compiler1.compile_default_action(batch_size)
        bounds = self.compiler1.compile_action_bound_constraints(initial_state)
        self.assertIsInstance(bounds, dict)
        self.assertIn('outflow/1', bounds)
        self.assertIsInstance(bounds['outflow/1'], tuple)
        self.assertEqual(len(bounds['outflow/1']), 2)
        lower, upper = bounds['outflow/1']
        self.assertIsInstance(lower, TensorFluent)
        self.assertListEqual(lower.shape.as_list(), [])
        self.assertEqual(lower.dtype, tf.float32)
        self.assertIsInstance(upper, TensorFluent)
        self.assertEqual(upper.dtype, tf.float32)
        self.assertListEqual(upper.shape.as_list(), [batch_size, 8])

    def test_instantiate_non_fluents(self):
        nf = dict(self.compiler1.non_fluents)

        expected_non_fluents = {
            'MAX_RES_CAP/1': { 'shape': [8,], 'dtype': tf.float32 },
            'UPPER_BOUND/1': { 'shape': [8,], 'dtype': tf.float32 },
            'LOWER_BOUND/1': { 'shape': [8,], 'dtype': tf.float32 },
            'RAIN_SHAPE/1': { 'shape': [8,], 'dtype': tf.float32 },
            'RAIN_SCALE/1': { 'shape': [8,], 'dtype': tf.float32 },
            'DOWNSTREAM/2': { 'shape': [8,8], 'dtype': tf.bool },
            'SINK_RES/1': { 'shape': [8,], 'dtype': tf.bool },
            'MAX_WATER_EVAP_FRAC_PER_TIME_UNIT/0': { 'shape': [], 'dtype': tf.float32 },
            'LOW_PENALTY/1': { 'shape': [8,], 'dtype': tf.float32 },
            'HIGH_PENALTY/1': { 'shape': [8,], 'dtype': tf.float32 }
        }
        self.assertIsInstance(nf, dict)
        self.assertEqual(len(nf), len(expected_non_fluents))
        for name, fluent in nf.items():
            self.assertIn(name, expected_non_fluents)
            shape = expected_non_fluents[name]['shape']
            dtype = expected_non_fluents[name]['dtype']
            self.assertEqual(fluent.name, '{}:0'.format(name))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'MAX_RES_CAP/1': [ 100.,  100.,  200.,  300.,  400.,  500.,  800., 1000.],
            'UPPER_BOUND/1': [ 80.,  80., 180., 280., 380., 480., 780., 980.],
            'LOWER_BOUND/1': [20., 20., 20., 20., 20., 20., 20., 20.],
            'RAIN_SHAPE/1': [1., 1., 1., 1., 1., 1., 1., 1.],
            'RAIN_SCALE/1': [ 5.,  3.,  9.,  7., 15., 13., 25., 30.],
            'DOWNSTREAM/2': [
                [False, False, False, False, False, True, False, False],
                [False, False, True, False, False, False, False, False],
                [False, False, False, False, True, False, False, False],
                [False, False, False, False, False, False, False, True],
                [False, False, False, False, False, False, True, False],
                [False, False, False, False, False, False, True, False],
                [False, False, False, False, False, False, False, True],
                [False, False, False, False, False, False, False, False]
            ],
            'SINK_RES/1': [False, False, False, False, False, False, False, True],
            'MAX_WATER_EVAP_FRAC_PER_TIME_UNIT/0': 0.05,
            'LOW_PENALTY/1': [-5., -5., -5., -5., -5., -5., -5., -5.],
            'HIGH_PENALTY/1': [-10., -10., -10., -10., -10., -10., -10., -10.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in nf.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_instantiate_initial_state_fluents(self):
        sf = dict(self.compiler1.initial_state_fluents)

        expected_state_fluents = {
            'rlevel/1': { 'shape': [8,] , 'dtype': tf.float32 }
        }
        self.assertIsInstance(sf, dict)
        self.assertEqual(len(sf), len(expected_state_fluents))
        for name, fluent in sf.items():
            self.assertIn(name, expected_state_fluents)
            shape = expected_state_fluents[name]['shape']
            dtype = expected_state_fluents[name]['dtype']
            self.assertEqual(fluent.name, '{}:0'.format(name))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'rlevel/1': [75., 50., 50., 50., 50., 50., 50., 50.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in sf.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_instantiate_default_action_fluents(self):
        action_fluents = self.compiler1.default_action_fluents
        self.assertIsInstance(action_fluents, list)
        for fluent in action_fluents:
            self.assertIsInstance(fluent, tuple)
            self.assertEqual(len(fluent), 2)
            self.assertIsInstance(fluent[0], str)
            self.assertIsInstance(fluent[1], TensorFluent)

        af = dict(action_fluents)

        expected_action_fluents = {
            'outflow/1': { 'shape': [8,] , 'dtype': tf.float32 }
        }
        self.assertEqual(len(af), len(expected_action_fluents))
        for name, fluent in af.items():
            self.assertIn(name, expected_action_fluents)
            shape = expected_action_fluents[name]['shape']
            dtype = expected_action_fluents[name]['dtype']
            self.assertEqual(fluent.name, '{}:0'.format(name))
            self.assertIsInstance(fluent, TensorFluent)
            self.assertEqual(fluent.dtype, dtype)
            self.assertEqual(fluent.shape.as_list(), shape)

        expected_initializers = {
            'outflow/1': [0., 0., 0., 0., 0., 0., 0., 0.]
        }
        with tf.Session(graph=self.compiler1.graph) as sess:
            for name, fluent in af.items():
                value = sess.run(fluent.tensor)
                list1 = list(value.flatten())
                list2 = list(np.array(expected_initializers[name]).flatten())
                for v1, v2 in zip(list1, list2):
                    self.assertAlmostEqual(v1, v2)

    def test_non_fluent_ordering(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            non_fluents = dict(compiler.non_fluents)
            action_fluent_ordering = compiler.non_fluent_ordering
            self.assertEqual(len(action_fluent_ordering), len(non_fluents))
            for action_fluent in action_fluent_ordering:
                self.assertIn(action_fluent, non_fluents)

    def test_state_fluent_ordering(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            initial_state_fluents = dict(compiler.initial_state_fluents)
            current_state_ordering = compiler.state_fluent_ordering
            self.assertEqual(len(current_state_ordering), len(initial_state_fluents))
            for fluent in initial_state_fluents:
                self.assertIn(fluent, current_state_ordering)

            next_state_ordering = compiler.next_state_fluent_ordering
            self.assertEqual(len(current_state_ordering), len(next_state_ordering))

            for current_fluent, next_fluent in zip(current_state_ordering, next_state_ordering):
                self.assertEqual(utils.rename_state_fluent(current_fluent), next_fluent)
                self.assertEqual(utils.rename_next_state_fluent(next_fluent), current_fluent)

    def test_interm_fluent_ordering(self):
        compilers = [self.compiler1, self.compiler2]
        expected = [
            ['evaporated/1', 'overflow/1', 'rainfall/1', 'inflow/1'],
            []
        ]
        for compiler, expected_ordering in zip(compilers, expected):
            interm_fluent_ordering = compiler.interm_fluent_ordering
            self.assertListEqual(interm_fluent_ordering, expected_ordering)

    def test_action_fluent_ordering(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            default_action_fluents = dict(compiler.default_action_fluents)
            action_fluent_ordering = compiler.action_fluent_ordering
            self.assertEqual(len(action_fluent_ordering), len(default_action_fluents))
            for action_fluent in action_fluent_ordering:
                self.assertIn(action_fluent, default_action_fluents)

    def test_state_fluent_variables(self):
        compilers = [self.compiler1, self.compiler2]
        fluent_variables = [
            {
            'rlevel/1': ['rlevel(t1)', 'rlevel(t2)', 'rlevel(t3)', 'rlevel(t4)', 'rlevel(t5)', 'rlevel(t6)', 'rlevel(t7)', 'rlevel(t8)']
            },
            {
                'picTaken/1': ['picTaken(p1)', 'picTaken(p2)', 'picTaken(p3)'],
                'time/0': ['time'],
                'xPos/0': ['xPos'],
                'yPos/0': ['yPos']
            }
        ]
        for compiler, expected_variables in zip(compilers, fluent_variables):
            fluent_variables = compiler.state_fluent_variables
            self.assertEqual(len(fluent_variables), len(expected_variables))
            for name, actual_variables in fluent_variables:
                self.assertIn(name, expected_variables)
                self.assertListEqual(actual_variables, expected_variables[name])

    def test_interm_fluent_variables(self):
        compilers = [self.compiler1, self.compiler2]
        fluent_variables = [
            {
            'evaporated/1': ['evaporated(t1)', 'evaporated(t2)', 'evaporated(t3)', 'evaporated(t4)', 'evaporated(t5)', 'evaporated(t6)', 'evaporated(t7)', 'evaporated(t8)'],
            'rainfall/1': ['rainfall(t1)', 'rainfall(t2)', 'rainfall(t3)', 'rainfall(t4)', 'rainfall(t5)', 'rainfall(t6)', 'rainfall(t7)', 'rainfall(t8)'],
            'overflow/1': ['overflow(t1)', 'overflow(t2)', 'overflow(t3)', 'overflow(t4)', 'overflow(t5)', 'overflow(t6)', 'overflow(t7)', 'overflow(t8)'],
            'inflow/1': ['inflow(t1)', 'inflow(t2)', 'inflow(t3)', 'inflow(t4)', 'inflow(t5)', 'inflow(t6)', 'inflow(t7)', 'inflow(t8)']
            },
            {}
        ]
        for compiler, expected_variables in zip(compilers, fluent_variables):
            fluent_variables = compiler.interm_fluent_variables
            self.assertEqual(len(fluent_variables), len(expected_variables))
            for name, actual_variables in fluent_variables:
                self.assertIn(name, expected_variables)
                self.assertListEqual(actual_variables, expected_variables[name])

    def test_action_fluent_variables(self):
        compilers = [self.compiler1, self.compiler2]
        fluent_variables = [
            {
                'outflow/1': ['outflow(t1)', 'outflow(t2)', 'outflow(t3)', 'outflow(t4)', 'outflow(t5)', 'outflow(t6)', 'outflow(t7)', 'outflow(t8)']
            },
            {
                'snapPicture/0': ['snapPicture'],
                'xMove/0': ['xMove'],
                'yMove/0': ['yMove']
            }
        ]
        for compiler, expected_variables in zip(compilers, fluent_variables):
            fluent_variables = compiler.action_fluent_variables
            self.assertEqual(len(fluent_variables), len(expected_variables))
            for name, actual_variables in fluent_variables:
                self.assertIn(name, expected_variables)
                self.assertListEqual(actual_variables, expected_variables[name])

    def test_state_size(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            state_size = compiler.state_size
            initial_state_fluents = dict(compiler.initial_state_fluents)
            state_fluent_ordering = compiler.state_fluent_ordering
            next_state_fluent_ordering = compiler.next_state_fluent_ordering

            self.assertIsInstance(state_size, tuple)
            for shape in state_size:
                self.assertIsInstance(shape, tuple)
            self.assertEqual(len(state_size), len(initial_state_fluents))
            self.assertEqual(len(state_size), len(state_fluent_ordering))
            self.assertEqual(len(state_size), len(next_state_fluent_ordering))

            for shape, name in zip(state_size, state_fluent_ordering):
                actual = list(shape)
                expected = initial_state_fluents[name].shape.as_list()
                if expected == []:
                    expected = [1]
                self.assertListEqual(actual, expected)

            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = {}
            scope.update(nf)
            scope.update(sf)
            scope.update(af)
            interm_fluents, next_state_fluents = compiler.compile_cpfs(scope)
            next_state_fluents = dict(next_state_fluents)
            for shape, name in zip(state_size, next_state_fluent_ordering):
                actual = list(shape)
                expected = next_state_fluents[name].shape.as_list()
                if expected == []:
                    expected = [1]
                self.assertListEqual(actual, expected)

    def test_action_size(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            action_size = compiler.action_size
            default_action_fluents = dict(compiler.default_action_fluents)
            action_fluent_ordering = compiler.action_fluent_ordering
            self.assertIsInstance(action_size, tuple)
            self.assertEqual(len(action_size), len(default_action_fluents))
            self.assertEqual(len(action_size), len(action_fluent_ordering))
            for shape in action_size:
                self.assertIsInstance(shape, tuple)

    def test_interm_size(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            interm_size = compiler.interm_size
            interm_ordering = compiler.interm_fluent_ordering
            self.assertIsInstance(interm_size, tuple)
            self.assertEqual(len(interm_size), len(interm_ordering))
            for shape in interm_size:
                self.assertIsInstance(shape, tuple)

    def test_state_dtype(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            state_dtype = compiler.state_dtype
            initial_state_fluents = compiler.initial_state_fluents
            state_fluent_ordering = compiler.state_fluent_ordering
            self.assertIsInstance(state_dtype, tuple)
            self.assertEqual(len(state_dtype), len(initial_state_fluents))
            self.assertEqual(len(state_dtype), len(state_fluent_ordering))
            for i, dtype in enumerate(state_dtype):
                self.assertIsInstance(dtype, tf.DType)
                self.assertEqual(dtype, initial_state_fluents[i][1].dtype)

    def test_interm_dtype(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            interm_dtype = compiler.interm_dtype
            interm_ordering = compiler.interm_fluent_ordering
            self.assertIsInstance(interm_dtype, tuple)
            self.assertEqual(len(interm_dtype), len(interm_ordering))
            for i, dtype in enumerate(interm_dtype):
                self.assertIsInstance(dtype, tf.DType)

    def test_action_dtype(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            action_dtype = compiler.action_dtype
            default_action_fluents = compiler.default_action_fluents
            action_fluent_ordering = compiler.action_fluent_ordering
            self.assertIsInstance(action_dtype, tuple)
            self.assertEqual(len(action_dtype), len(default_action_fluents))
            self.assertEqual(len(action_dtype), len(action_fluent_ordering))
            for i, dtype in enumerate(action_dtype):
                self.assertIsInstance(dtype, tf.DType)
                self.assertEqual(dtype, default_action_fluents[i][1].dtype)

    def test_state_scope(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            fluents = compiler.initial_state_fluents
            scope = dict(fluents)
            self.assertEqual(len(fluents), len(scope))
            for i, name in enumerate(compiler.state_fluent_ordering):
                self.assertIs(scope[name], fluents[i][1])

    def test_state_scope(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            fluents = compiler.default_action_fluents
            scope = dict(fluents)
            self.assertEqual(len(fluents), len(scope))
            for i, name in enumerate(compiler.action_fluent_ordering):
                self.assertIs(scope[name], fluents[i][1])

    def test_compile_expressions(self):
        expected = {
            # rddl1: RESERVOIR ====================================================
            'rainfall/1':   { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'evaporated/1': { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'overflow/1':   { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            'inflow/1':     { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },
            "rlevel'/1":    { 'shape': [8,], 'dtype': tf.float32, 'scope': ['?r'] },

            # rddl2: MARS ROVER ===================================================
            "xPos'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "yPos'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "time'/0":   { 'shape': [], 'dtype': tf.float32, 'scope': [] },
            "picTaken'/1": { 'shape': [3,], 'dtype': tf.bool, 'scope': ['?p'] }
        }

        compilers = [self.compiler1, self.compiler2]
        rddls = [self.rddl1, self.rddl2]
        for compiler, rddl in zip(compilers, rddls):
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = {}
            scope.update(nf)
            scope.update(sf)
            scope.update(af)

            _, cpfs = rddl.domain.cpfs
            for cpf in cpfs:
                name = cpf.name
                expr = cpf.expr
                t = compiler._compile_expression(expr, scope)
                scope[name] = t
                self.assertIsInstance(t, TensorFluent)
                self.assertEqual(t.shape.as_list(), expected[name]['shape'])
                self.assertEqual(t.dtype, expected[name]['dtype'])
                self.assertEqual(t.scope.as_list(), expected[name]['scope'])

            reward_expr = rddl.domain.reward
            t = compiler._compile_expression(reward_expr, scope)
            self.assertIsInstance(t, TensorFluent)
            self.assertEqual(t.shape.as_list(), [])
            self.assertEqual(t.dtype, tf.float32)
            self.assertEqual(t.scope.as_list(), [])

    def test_compile_cpfs(self):
        compilers = [self.compiler1, self.compiler2]
        expected = [
            (['evaporated/1', 'overflow/1', 'rainfall/1', 'inflow/1'], ["rlevel'/1"]),
            ([], ["picTaken'/1", "time'/0", "xPos'/0", "yPos'/0"]),
        ]
        for compiler, (expected_interm, expected_state) in zip(compilers, expected):
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = { **nf, **sf, **af }

            interm_fluents, next_state_fluents = compiler.compile_cpfs(scope)

            self.assertIsInstance(interm_fluents, list)
            self.assertEqual(len(interm_fluents), len(expected_interm))
            for fluent, expected_fluent in zip(interm_fluents, expected_interm):
                self.assertEqual(fluent[0], expected_fluent)

            self.assertIsInstance(next_state_fluents, list)
            self.assertEqual(len(next_state_fluents), len(expected_state))
            for fluent, expected_fluent in zip(next_state_fluents, expected_state):
                self.assertEqual(fluent[0], expected_fluent)

    def test_compile_state_cpfs(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = { **nf, **sf, **af }

            interm_fluents = compiler.compile_intermediate_cpfs(scope)
            scope.update(dict(interm_fluents))
            next_state_fluents = compiler.compile_state_cpfs(scope)

            self.assertIsInstance(next_state_fluents, list)
            for cpf in next_state_fluents:
                self.assertIsInstance(cpf, tuple)
            self.assertEqual(len(next_state_fluents), len(sf))

            next_state_fluents = dict(next_state_fluents)
            for fluent in sf:
                next_fluent = utils.rename_state_fluent(fluent)
                self.assertIn(next_fluent, next_state_fluents)
                self.assertIsInstance(next_state_fluents[next_fluent], TensorFluent)

    def test_compile_intermediate_cpfs(self):
        compilers = [self.compiler1, self.compiler2]
        expected = [
            ['evaporated/1', 'overflow/1', 'rainfall/1', 'inflow/1'],
            []
        ]
        for compiler, fluents in zip(compilers, expected):
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = { **nf, **sf, **af }
            interm_fluents = compiler.compile_intermediate_cpfs(scope)
            self.assertIsInstance(interm_fluents, list)
            self.assertEqual(len(interm_fluents), len(fluents))
            for actual, expected in zip(interm_fluents, fluents):
                self.assertIsInstance(actual, tuple)
                self.assertEqual(len(actual), 2)
                self.assertIsInstance(actual[0], str)
                self.assertIsInstance(actual[1], TensorFluent)
                self.assertEqual(actual[0], expected)

    def test_compile_reward(self):
        compilers = [self.compiler1, self.compiler2]
        for compiler in compilers:
            nf = compiler.non_fluents_scope()
            sf = dict(compiler.initial_state_fluents)
            af = dict(compiler.default_action_fluents)
            scope = { **nf, **sf, **af }
            interm_fluents, next_state_fluents = compiler.compile_cpfs(scope)
            scope.update(dict(next_state_fluents))
            reward = compiler.compile_reward(scope)
            self.assertIsInstance(reward, TensorFluent)
            self.assertEqual(reward.shape.as_list(), [1])



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.


from typing import List, Iterator, Tuple

ParamsList = List[str]
Permutation = List[int]
BroadcastTuple = Tuple[ParamsList, Permutation, Permutation]


class TensorFluentScope(object):
    '''TensorFluentScope handles the ordering of the fluent's parameters.

    Each RDDL fluent is defined by a parameterized variable.
    The TensorFluentScope class manages to match the order of the
    fluent's parameters with the tensor's shape when evaluating
    operations and functions in expressions.

    Args:
        scope: The list of the parameters of the fluent
    '''

    def __init__(self, scope: List[str]) -> None:
        self._scope = scope

    def as_list(self) -> List[str]:
        '''Returns a copy of the fluent's scope.'''
        return self._scope[:]

    def index(self, p: str) -> int:
        '''Returns the position of parameter `p` in the scope.'''
        return self._scope.index(p)

    def __len__(self) -> int:
        '''Returns the number of parameters in the scope.'''
        return len(self._scope)

    def __getitem__(self, i: int) -> str:
        '''Returns the parameter at position `i`.'''
        return self._scope[i]

    def __eq__(self, other: 'TensorFluentScope') -> bool:
        '''Returns True if scopes are equal. False, otherwise.'''
        return self._scope == other._scope

    def __ne__(self, other: 'TensorFluentScope') -> bool:
        '''Returns True if scopes are not equal. False, otherwise.'''
        return self._scope != other._scope

    def __str__(self) -> str:
        '''Returns the string representation of a TensorFluentScope object.'''
        return 'TensorFluentScope({})'.format(str(self._scope))

    @classmethod
    def broadcast(cls, s1: ParamsList, s2: ParamsList) -> BroadcastTuple:
        '''It broadcasts the smaller scope over the larger scope.

        It handles scope intersection as well as differences in scopes
        in order to output a resulting scope so that input scopes are
        contained within it (i.e., input scopes are subscopes of the
        output scope). Also, if necessary, it outputs permutations of
        the input scopes so that tensor broadcasting invariants are
        not violated.

        Note:
            For more information on broadcasting, please report to
            NumPy's official documentation available at the following URLs:
            1. https://docs.scipy.org/doc/numpy/user/basics.broadcasting.html
            2. https://docs.scipy.org/doc/numpy/reference/generated/numpy.broadcast.html

        Args:
            s1: A fluent's scope.
            s2: A fluent's scope.

        Returns:
            A tuple with the output scope and permutations of the input scopes.
        '''
        if len(s1) == 0:
            return s2, [], []
        if len(s2) == 0:
            return s1, [], []

        subscope = list(set(s1) & set(s2))
        if len(subscope) == len(s1):
            subscope = s1
        elif len(subscope) == len(s2):
            subscope = s2

        perm1 = []
        if s1[-len(subscope):] != subscope:
            i = 0
            for var in s1:
                if var not in subscope:
                    perm1.append(i)
                    i += 1
                else:
                    j = subscope.index(var)
                    perm1.append(len(s1) - len(subscope) + j)
        perm2 = []
        if s2[-len(subscope):] != subscope:
            i = 0
            for var in s2:
                if var not in subscope:
                    perm2.append(i)
                    i += 1
                else:
                    j = subscope.index(var)
                    perm2.append(len(s2) - len(subscope) + j)

        scope = [] # type: ParamsList
        if len(s1) >= len(s2):
            if perm1 == []:
                scope = s1
            else:
                for i in range(len(s1)):
                    scope.append(s1[perm1.index(i)])
        else:
            if perm2 == []:
                scope = s2
            else:
                for i in range(len(s2)):
                    scope.append(s2[perm2.index(i)])

        return (scope, perm1, perm2)



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.


from tfrddlsim.rddl2tf.compiler import Compiler
from tfrddlsim.rddl2tf.fluent import TensorFluent
from tfrddlsim.policy import Policy

import numpy as np
import tensorflow as tf


from typing import Iterable, Sequence, Optional, Tuple

Shape = Sequence[int]
FluentPair = Tuple[str, TensorFluent]

NonFluentsTensor = Sequence[tf.Tensor]
StateTensor = Sequence[tf.Tensor]
StatesTensor = Sequence[tf.Tensor]
ActionsTensor = Sequence[tf.Tensor]
IntermsTensor = Sequence[tf.Tensor]

NonFluentsArray = Sequence[np.array]
StateArray = Sequence[np.array]
StatesArray = Sequence[np.array]
ActionsArray = Sequence[np.array]
IntermsArray = Sequence[np.array]

CellOutput = Tuple[StatesTensor, ActionsTensor, IntermsTensor, tf.Tensor]
CellState = Sequence[tf.Tensor]

TrajectoryOutput = Tuple[StateTensor, StatesTensor, ActionsTensor, IntermsTensor, tf.Tensor]
SimulationOutput = Tuple[NonFluentsArray, StateArray, StatesArray, ActionsArray, IntermsArray, np.array]


class PolicySimulationCell(tf.nn.rnn_cell.RNNCell):
    '''SimulationCell implements a 1-step MDP transition cell.

    It extends`tf.nn.rnn_cell.RNNCell` for simulating an MDP transition for a given policy.
    The cell input is the timestep. The hidden state is the factored MDP state.
    The cell output is the tuple of MDP fluents (next-state, action, interm, rewards).

    Note:
        All fluents are represented in factored form as Tuple[tf.Tensors].

    Args:
        compiler (:obj:`tfrddlsim.compiler.Compiler`): RDDL2TensorFlow compiler.
        policy (:obj:`tfrddlsim.policy.Policy`): MDP Policy.
        batch_size (int): The size of the simulation batch.
    '''

    def __init__(self, compiler: Compiler, policy: Policy, batch_size: int) -> None:
        self._compiler = compiler
        self._policy = policy
        self._batch_size = batch_size

    @property
    def graph(self) -> tf.Graph:
        '''Returns the computation graph.'''
        return self._compiler.graph

    @property
    def state_size(self) -> Sequence[Shape]:
        '''Returns the MDP state size.'''
        return self._compiler.state_size

    @property
    def action_size(self) -> Sequence[Shape]:
        '''Returns the MDP action size.'''
        return self._compiler.action_size

    @property
    def interm_size(self) -> Sequence[Shape]:
        '''Returns the MDP intermediate state size.'''
        return self._compiler.interm_size

    @property
    def output_size(self) -> Tuple[Sequence[Shape], Sequence[Shape], Sequence[Shape], int]:
        '''Returns the simulation cell output size.'''
        return (self.state_size, self.action_size, self.interm_size, 1)

    def initial_state(self) -> StateTensor:
        '''Returns the initial state tensor.'''
        return self._compiler.compile_initial_state(self._batch_size)

    def __call__(self,
            input: tf.Tensor,
            state: Sequence[tf.Tensor],
            scope: Optional[str] = None) -> Tuple[CellOutput, CellState]:
        '''Returns the simulation cell for the given `input` and `state`.

        The cell returns states, actions and interms as a
        sequence of tensors (i.e., all representations are factored).
        The reward is an 1-dimensional tensor.

        Note:
            All tensors have shape: (batch_size, fluent_shape).

        Args:
            input (tf.Tensor): The current MDP timestep.
            state (Sequence[tf.Tensor]): State fluents in canonical order.
            scope (Optional[str]): Scope for operations in graph.

        Returns:
            Tuple[CellOutput, CellState]: (output, next_state).
        '''
        # action
        action = self._policy(state, input)

        # next state
        transition_scope = self._compiler.transition_scope(state, action)
        interm_fluents, next_state_fluents = self._compiler.compile_cpfs(transition_scope, self._batch_size)
        next_state = tuple(self._tensors(next_state_fluents))

        # reward
        transition_scope.update(next_state_fluents)
        reward = self._compiler.compile_reward(transition_scope)
        reward = reward.tensor

        # outputs
        output_interm_state = self._output(interm_fluents)
        output_next_state = self._output(next_state_fluents)
        output_action = tuple(self._dtype(tensor) for tensor in action)
        output = (output_next_state, output_action, output_interm_state, reward)

        return (output, next_state)

    @classmethod
    def _tensors(cls, fluents: Sequence[FluentPair]) -> Iterable[tf.Tensor]:
        '''Yields the `fluents`' tensors.'''
        for _, fluent in fluents:
            yield fluent.tensor

    @classmethod
    def _dtype(cls, tensor: tf.Tensor) -> tf.Tensor:
        '''Converts `tensor` to tf.float32 datatype if needed.'''
        if tensor.dtype != tf.float32:
            tensor = tf.cast(tensor, tf.float32)
        return tensor

    @classmethod
    def _output(cls, fluents: Sequence[FluentPair]) -> Sequence[tf.Tensor]:
        '''Returns output tensors for `fluents`.'''
        return tuple(cls._dtype(t) for t in cls._tensors(fluents))


class PolicySimulator(object):
    '''Simulator class samples MDP trajectories in the computation graph.

    It implements the n-step MDP trajectory simulator using dynamic unrolling
    in a recurrent model. Its inputs are the MDP initial state and the number
    of timesteps in the horizon.

    Args:
        compiler (:obj:`tfrddlsim.compiler.Compiler`): RDDL2TensorFlow compiler.
        policy (:obj:`tfrddlsim.policy.Policy`): MDP Policy.
        batch_size (int): The size of the simulation batch.
    '''

    def __init__(self, compiler: Compiler, policy: Policy, batch_size: int) -> None:
        self._cell = PolicySimulationCell(compiler, policy, batch_size)
        self._non_fluents = [fluent.tensor for _, fluent in compiler.non_fluents]

    @property
    def graph(self):
        '''Returns the computation graph.'''
        return self._cell.graph

    @property
    def batch_size(self) -> int:
        '''Returns the size of the simulation batch.'''
        return self._cell._batch_size

    @property
    def input_size(self) -> int:
        '''Returns the simulation input size (e.g., timestep).'''
        return 1

    @property
    def state_size(self) -> Sequence[Shape]:
        '''Returns the MDP state size.'''
        return self._cell.state_size

    @property
    def output_size(self) -> Tuple[Sequence[Shape], Sequence[Shape], Sequence[Shape], int]:
        '''Returns the simulation output size.'''
        return self._cell.output_size

    def timesteps(self, horizon: int) -> tf.Tensor:
        '''Returns the input tensor for the given `horizon`.'''
        start, limit, delta = horizon - 1, -1, -1
        timesteps_range = tf.range(start, limit, delta, dtype=tf.float32)
        timesteps_range = tf.expand_dims(timesteps_range, -1)
        batch_timesteps = tf.stack([timesteps_range] * self.batch_size)
        return batch_timesteps

    def trajectory(self,
            horizon: int,
            initial_state: Optional[StateTensor] = None) -> TrajectoryOutput:
        '''Returns the ops for the trajectory generation with given `horizon`
        and `initial_state`.

        The simulation returns states, actions and interms as a
        sequence of tensors (i.e., all representations are factored).
        The reward is a batch sized tensor.
        The trajectoty output is a tuple: (initial_state, states, actions, interms, rewards).
        If initial state is None, use default compiler's initial state.

        Note:
            All tensors have shape: (batch_size, horizon, fluent_shape).
            Except initial state that has shape: (batch_size, fluent_shape).

        Args:
            horizon (int): The number of simulation timesteps.
            initial_state (Optional[Sequence[tf.Tensor]]): The initial state tensors.

        Returns:
            Tuple[StateTensor, StatesTensor, ActionsTensor, IntermsTensor, tf.Tensor]: Trajectory output tuple.
        '''
        if initial_state is None:
            initial_state = self._cell.initial_state()

        with self.graph.as_default():
            inputs = self.timesteps(horizon)
            outputs, _ = tf.nn.dynamic_rnn(
                                self._cell,
                                inputs,
                                initial_state=initial_state,
                                dtype=tf.float32,
                                scope="trajectory")
            states, actions, interms, rewards = outputs

            # fluent types
            state_dtype = self._cell._compiler.state_dtype
            states = self._output(states, state_dtype)
            interm_dtype = self._cell._compiler.interm_dtype
            interms = self._output(interms, interm_dtype)
            action_dtype = self._cell._compiler.action_dtype
            actions = self._output(actions, action_dtype)

            outputs = (initial_state, states, actions, interms, rewards)

        return outputs

    def run(self,
            horizon: int,
            initial_state: Optional[StateTensor] = None) -> SimulationOutput:
        '''Builds the MDP graph and simulates in batch the trajectories
        with given `horizon`. Returns the non-fluents, states, actions, interms
        and rewards. Fluents and non-fluents are returned in factored form.

        Note:
            All output arrays have shape: (batch_size, horizon, fluent_shape).
            Except initial state that has shape: (batch_size, fluent_shape).

        Args:
            horizon (int): The number of timesteps in the simulation.
            initial_state (Optional[Sequence[tf.Tensor]]): The initial state tensors.

        Returns:
            Tuple[NonFluentsArray, StatesArray, ActionsArray, IntermsArray, np.array]: Simulation ouput tuple.
        '''
        trajectory = self.trajectory(horizon, initial_state)

        with tf.Session(graph=self.graph) as sess:
            sess.run(tf.global_variables_initializer())
            non_fluents = sess.run(self._non_fluents)
            initial_state, states, actions, interms, rewards = sess.run(trajectory)

        # non-fluents
        non_fluent_ordering = self._cell._compiler.non_fluent_ordering
        non_fluents = tuple(zip(non_fluent_ordering, non_fluents))

        # states
        state_fluent_ordering = self._cell._compiler.state_fluent_ordering
        states = tuple(zip(state_fluent_ordering, states))

        # interms
        interm_fluent_ordering = self._cell._compiler.interm_fluent_ordering
        interms = tuple(zip(interm_fluent_ordering, interms))

        # actions
        action_fluent_ordering = self._cell._compiler.action_fluent_ordering
        actions = tuple(zip(action_fluent_ordering, actions))

        # rewards
        rewards = np.squeeze(rewards)

        outputs = (non_fluents, initial_state, states, actions, interms, rewards)

        return outputs

    @classmethod
    def _output(cls,
            tensors: Sequence[tf.Tensor],
            dtypes: Sequence[tf.DType]) -> Sequence[tf.Tensor]:
        '''Converts `tensors` to the corresponding `dtypes`.'''
        outputs = []
        for tensor, dtype in zip(tensors, dtypes):
            tensor = tensor[0]
            if tensor.dtype != dtype:
                tensor = tf.cast(tensor, dtype)
            outputs.append(tensor)
        return tuple(outputs)



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.


import tensorflow as tf

from typing import List, Optional, Sequence, Tuple

Reshaping = Optional[List[int]]


class TensorFluentShape(object):
    '''TensorFluentShape manages the fluent's shape in operations.

    Each RDDL fluent is a parameterized variable. The arity and
    the number of objects corresponding to the type of each parameter
    of a fluent are reflected in a ``tfrddlsim.TensorFluentShape``
    object (i.e., the rank of a ``tfrddlsim.TensorFluent`` corresponds
    to the fluent arity and the size of its dimensions corresponds
    to the number of objects of each type).

    Also, a ``tfrddlsim.TensorFluentShape`` manages batch sizes
    when evaluating operations in batch mode.

    If in batch mode, the first dimension always corresponds
    to the batch size.

    Args:
        shape: The list of dimensions of the fluent.
        batch: The batch mode flag.
    '''

    def __init__(self, shape: List[int], batch: bool) -> None:
        self._shape = tf.TensorShape(shape)
        self._batch = batch

    def as_list(self) -> List[str]:
        '''Returns the fluent's shape as a list.'''
        return self._shape.as_list()

    def __getitem__(self, i: int) -> int:
        '''Returns the size of dimension `i`.'''
        return self._shape[i]

    def __eq__(self, other: 'TensorFluentShape') -> bool:
        '''Returns True if fluent shapes are equal. False, otherwise.'''
        return self._shape == other._shape and self._batch == other._batch

    def __ne__(self, other: 'TensorFluentShape') -> bool:
        '''Returns True if fluent shapes are not equal. False, otherwise.'''
        return self._shape != other._shape or self._batch != other._batch

    def __str__(self) -> str:
        '''Returns the string represenation of a TensorFluentShape object.'''
        return 'TensorFluentShape({}, batch={})'.format(self.as_list(), self._batch)

    @property
    def batch(self) -> bool:
        '''Returns True if fluent is in batch mode. False, otherwise.'''
        return self._batch

    @property
    def batch_size(self) -> int:
        '''Returns the batch size if in batch mode. Otherwise, returns 1.'''
        return self._shape.as_list()[0] if self._batch else 1

    @property
    def fluent_shape(self) -> Sequence[int]:
        '''Returns a copy of the fluent shape, ignoring batch size if in batch mode.'''
        return tuple(self._shape.as_list()[1:] if self._batch else self._shape.as_list()[:])

    @property
    def fluent_size(self) -> int:
        '''Returns the size of fluent shape (i.e., the fluent's arity).'''
        return len(self.fluent_shape)

    @classmethod
    def broadcast(cls,
            shape1: 'TensorFluentShape',
            shape2: 'TensorFluentShape') -> Tuple[Reshaping, Reshaping]:
        '''It broadcasts the fluent shapes if any input is in batch mode.

        It handles input shapes in different modes, expanding its
        dimensions if necessary. It outputs a tuple with new shapes.
        If no input shape is in batch mode, return (None, None).
        If an input shape does not need to be changed, return None.

        Args:
            shape1: A fluent's shape.
            shape2: A fluent's shape.

        Returns:
            A pair of new shapes.
        '''
        reshape_1, reshape_2 = None, None

        if not (shape1._batch or shape2._batch):
            return reshape_1, reshape_2

        size_1, size_2 = shape1.fluent_size, shape2.fluent_size
        size_diff = abs(size_1 - size_2)
        if size_diff == 0:
            return reshape_1, reshape_2

        if size_2 > size_1 and not (size_1 == 0 and not shape1._batch):
            reshape_1 = [1] * size_diff + list(shape1.fluent_shape)
            if shape1._batch:
                reshape_1 = [shape1.batch_size] + reshape_1
        elif size_1 > size_2 and not (size_2 == 0 and not shape2._batch):
            reshape_2 = [1] * size_diff + list(shape2.fluent_shape)
            if shape2._batch:
                reshape_2 = [shape2.batch_size] + reshape_2
        return reshape_1, reshape_2



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.


from pyrddl.rddl import RDDL
from pyrddl.pvariable import PVariable
from pyrddl.expr import Expression

from tfrddlsim.rddl2tf.fluent import TensorFluent

import itertools
import numpy as np
import tensorflow as tf

from typing import Dict, List, Optional, Sequence, Tuple, Union

CPFPair = Tuple[str, TensorFluent]
FluentList = List[Tuple[str, TensorFluent]]
Bounds = Tuple[Optional[TensorFluent], Optional[TensorFluent]]
ObjectStruct = Dict[str, Union[int, Dict[str, int], List[str]]]
ObjectTable = Dict[str, ObjectStruct]
FluentParamsList = Sequence[Tuple[str, List[str]]]
Value = Union[bool, int, float]
ArgsList = Optional[List[str]]
InitializerPair = Tuple[Tuple[str, ArgsList], Value]
InitializerList = List[InitializerPair]


class Compiler(object):
    '''RDDL2TensorFlow compiler.

    This is the core component of tf-rddlsim package.

    Its API provides methods to compile RDDL fluents and expressions
    to TensorFlow tensors wrapped as :obj:`tfrddlsim.fluent.TensorFluent` objects.
    It supports constants, random variables, functions and operators
    used in most RDDL expressions. Also, it can handle next state
    and intermediate fluent CPFs, and rewards and action constraints.

    Args:
        rddl (:obj:`tfrddlsim.rddl.RDDL`): The RDDL model.
        batch_mode (bool): The batch mode flag.

    Attributes:
        rddl (:obj:`tfrddlsim.rddl.RDDL`): The RDDL model.
        batch_mode (bool): The batch mode flag.
        graph (:obj:`tensorflow.python.framework.ops.Graph`): The computation graph.
    '''

    def __init__(self, rddl: RDDL, batch_mode: bool = False) -> None:
        self.rddl = rddl
        self.batch_mode = batch_mode
        self.graph = tf.Graph()

    def batch_mode_on(self):
        '''Sets on the batch mode flag.'''
        self.batch_mode = True

    def batch_mode_off(self):
        '''Sets off the batch mode flag.'''
        self.batch_mode = False

    def compile_initial_state(self, batch_size: int) -> Sequence[tf.Tensor]:
        '''Returns a tuple of tensors representing the initial state fluents.

        Args:
            batch_size (int): The batch size.

        Returns:
            A tuple of tensors.
        '''
        return self._compile_batch_fluents(self.initial_state_fluents, batch_size)

    def compile_default_action(self, batch_size: int) -> Sequence[tf.Tensor]:
        '''Returns a tuple of tensors representing the default action fluents.

        Args:
            batch_size (int): The batch size.

        Returns:
            A tuple of tensors.
        '''
        return self._compile_batch_fluents(self.default_action_fluents, batch_size)

    def compile_cpfs(self,
            scope: Dict[str, TensorFluent],
            batch_size: Optional[int] = None) -> Tuple[List[CPFPair], List[CPFPair]]:
        '''Compiles the intermediate and next state fluent CPFs.

        Args:
            scope (Dict[str, :obj:`tfrddlsim.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            Tuple[List[CPFPair], List[CPFPair]]: A pair of lists of TensorFluent
            representing the intermediate and state CPFs.
        '''
        interm_fluents = self.compile_intermediate_cpfs(scope, batch_size)
        scope.update(dict(interm_fluents))
        next_state_fluents = self.compile_state_cpfs(scope, batch_size)
        return interm_fluents, next_state_fluents

    def compile_intermediate_cpfs(self,
            scope: Dict[str, TensorFluent],
            batch_size: Optional[int] = None) -> Tuple[List[CPFPair], List[CPFPair]]:
        '''Compiles the intermediate fluent CPFs.

        Args:
            scope (Dict[str, :obj:`tfrddlsim.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            A list of intermediate fluent CPFs compiled to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        interm_fluents = []
        for cpf in self.rddl.domain.intermediate_cpfs:
            t = self._compile_expression(cpf.expr, scope, batch_size)
            interm_fluents.append((cpf.name, t))
            scope[cpf.name] = t
        return interm_fluents

    def compile_state_cpfs(self,
            scope: Dict[str, TensorFluent],
            batch_size: Optional[int] = None) -> Tuple[List[CPFPair], List[CPFPair]]:
        '''Compiles the next state fluent CPFs given the current `state` and `action` scope.

        Args:
            scope (Dict[str, :obj:`tfrddlsim.fluent.TensorFluent`]): The fluent scope for CPF evaluation.
            batch_size (Optional[int]): The batch size.

        Returns:
            A list of state fluent CPFs compiled to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        next_state_fluents = []
        for cpf in self.rddl.domain.state_cpfs:
            t = self._compile_expression(cpf.expr, scope, batch_size)
            next_state_fluents.append((cpf.name, t))
        key = lambda f: self.next_state_fluent_ordering.index(f[0])
        next_state_fluents = sorted(next_state_fluents, key=key)
        return next_state_fluents

    def compile_reward(self, scope: Dict[str, TensorFluent]) -> TensorFluent:
        '''Compiles the reward function given the fluent `scope`.

        Args:
            scope (Dict[str, :obj:`tfrddlsim.fluent.TensorFluent`]): The fluent scope for reward evaluation.

        Returns:
            A :obj:`tfrddlsim.fluent.TensorFluent` representing the reward function.
        '''
        reward_expr = self.rddl.domain.reward
        t = self._compile_expression(reward_expr, scope)
        tensor = tf.expand_dims(t.tensor, -1)
        return TensorFluent(tensor, t.scope[:], t.batch)

    def compile_action_preconditions(self,
            state: Sequence[tf.Tensor],
            action: Sequence[tf.Tensor]) -> List[TensorFluent]:
        '''Compiles the action preconditions given current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A list of :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        scope = self.action_precondition_scope(state, action)
        preconds = []
        for p in self.action_preconditions:
            t = self._compile_expression(p, scope)
            tensor = t.tensor
            if t.shape.fluent_shape == ():
                tensor = tf.expand_dims(tensor, -1)
            fluent = TensorFluent(tensor, t.scope[:], t.batch)
            preconds.append(fluent)
        return preconds

    def compile_action_preconditions_checking(self,
            state: Sequence[tf.Tensor],
            action: Sequence[tf.Tensor]) -> tf.Tensor:
        '''Combines the action preconditions into an applicability checking op.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A boolean tensor for checking if `action` is application in `state`.
        '''
        preconds = self.compile_action_preconditions(state, action)
        all_preconds = tf.concat([p.tensor for p in preconds], axis=1)
        checking = tf.reduce_all(all_preconds, axis=1)
        return checking

    def compile_action_bound_constraints(self,
            state: Sequence[tf.Tensor]) -> Dict[str, Bounds]:
        '''Compiles all actions bounds for the given `state`.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A mapping from action names to a pair of
            :obj:`tfrddlsim.fluent.TensorFluent` representing
            its lower and upper bounds.
        '''
        scope = self.action_precondition_scope(state)

        lower_bounds = self.action_lower_bound_constraints
        upper_bounds = self.action_upper_bound_constraints

        bounds = {}
        for name in self.action_fluent_ordering:

            lower_expr = lower_bounds.get(name)
            lower = None
            if lower_expr is not None:
                lower = self._compile_expression(lower_expr, scope)

            upper_expr = upper_bounds.get(name)
            upper = None
            if upper_expr is not None:
                upper = self._compile_expression(upper_expr, scope)

            bounds[name] = (lower, upper)

        return bounds

    def non_fluents_scope(self) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with non-fluents.'''
        return dict(self.non_fluents)

    def state_scope(self, state_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current state-fluents.

        Args:
            state_fluents (Sequence[tf.Tensor]): The current state fluents.

        Returns:
            A mapping from state fluent names to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        return dict(zip(self.state_fluent_ordering, state_fluents))

    def action_scope(self, action_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current action-fluents.

        Args:
            action_fluents (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from action fluent names to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        return dict(zip(self.action_fluent_ordering, action_fluents))

    def next_state_scope(self, next_state_fluents: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns a partial scope with current next state-fluents.

        Args:
            next_state_fluents (Sequence[tf.Tensor]): The next state fluents.

        Returns:
            A mapping from next state fluent names to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        return dict(zip(self.next_state_fluent_ordering, next_state_fluents))

    def transition_scope(self,
        state: Sequence[tf.Tensor],
        action: Sequence[tf.Tensor]) -> Dict[str, TensorFluent]:
        '''Returns the complete transition fluent scope
        for the current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from fluent names to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        scope.update(self.action_scope(action))
        return scope

    def action_precondition_scope(self,
            state: Sequence[tf.Tensor],
            action: Optional[Sequence[tf.Tensor]] = None) -> Dict[str, TensorFluent]:
        '''Returns the action precondition fluent scope
        for the current `state` and `action` fluents.

        Args:
            state (Sequence[tf.Tensor]): The current state fluents.
            action (Sequence[tf.Tensor]): The action fluents.

        Returns:
            A mapping from fluent names to :obj:`tfrddlsim.fluent.TensorFluent`.
        '''
        scope = {}
        scope.update(self.non_fluents_scope())
        scope.update(self.state_scope(state))
        if action is not None:
            scope.update(self.action_scope(action))
        return scope

    @property
    def object_table(self) -> ObjectTable:
        '''The object table for each RDDL type.

        Returns:
            A mapping from type name to the type size,
            objects index and objects list.
        '''
        if self.__dict__.get('_object_table') is None:
            self._build_object_table()
        return self._object_table

    @property
    def non_fluents(self) -> FluentList:
        '''The list of non-fluents instantiated for a given RDDL non-fluents.

        Returns:
            List[Tuple[str, TensorFluent]]: the list of non-fluents.
        '''
        if self.__dict__.get('_non_fluents') is None:
            self._instantiate_non_fluents()
        return self._non_fluents

    @property
    def initial_state_fluents(self) -> FluentList:
        '''The list of initial state-fluents instantiated for a given RDDL instance.

        Returns:
            List[Tuple[str, TensorFluent]]: the list of state fluents.
        '''
        if self.__dict__.get('_initial_state_fluents') is None:
            self._instantiate_initial_state_fluents()
        return self._initial_state_fluents

    @property
    def default_action_fluents(self) -> FluentList:
        '''The list of non-fluents instantiated for a given RDDL domain.

        Returns:
            List[Tuple[str, TensorFluent]]: the list of action fluents.
        '''
        if self.__dict__.get('_default_action_fluents') is None:
            self._instantiate_default_action_fluents()
        return self._default_action_fluents

    @property
    def action_preconditions(self) -> Dict[str, List[Expression]]:
        '''The action precondition expressions.

        Returns:
            Dict[str, List[Expression]]: A mapping from fluent name to a list of Expressions.'''
        return self.rddl.domain.preconds

    @property
    def local_action_preconditions(self) -> Dict[str, List[Expression]]:
        '''The local action precondition expressions.

        Returns:
            Dict[str, List[Expression]]: A mapping from fluent name to a list of Expressions.'''
        if self.__dict__.get('_local_action_preconditions') is None:
            self._build_preconditions_table()
        return self._local_action_preconditions

    @property
    def global_action_preconditions(self) -> Dict[str, List[Expression]]:
        '''The global action precondition expressions.

        Returns:
            Dict[str, List[Expression]]: A mapping from fluent name to a list of Expressions.'''
        if self.__dict__.get('_global_action_preconditions') is None:
            self._build_preconditions_table()
        return self._global_action_preconditions

    @property
    def action_lower_bound_constraints(self) -> Dict[str, Expression]:
        '''The action lower bound constraint expressions.

        Returns:
            Dict[str, Expression]: A mapping from fluent name to an Expression.'''
        if self.__dict__.get('_action_lower_bound_constraints') is None:
            self._build_action_bound_constraints_table()
        return self._action_lower_bound_constraints

    @property
    def action_upper_bound_constraints(self) -> Dict[str, Expression]:
        '''The action upper bound constraint expressions.

        Returns:
            Dict[str, Expression]: A mapping from fluent name to an Expression.'''
        if self.__dict__.get('_action_upper_bound_constraints') is None:
            self._build_action_bound_constraints_table()
        return self._action_upper_bound_constraints

    @property
    def non_fluent_ordering(self) -> List[str]:
        '''The list of non-fluent names in canonical order.

        Returns:
            List[str]: A list of fluent names.
        '''
        return [name for name in sorted(self.rddl.domain.non_fluents)]

    @property
    def state_fluent_ordering(self) -> List[str]:
        '''The list of state-fluent names in canonical order.

        Returns:
            List[str]: A list of fluent names.
        '''
        return [name for name in sorted(self.rddl.domain.state_fluents)]

    @property
    def action_fluent_ordering(self) -> List[str]:
        '''The list of action-fluent names in canonical order.

        Returns:
            List[str]: A list of fluent names.
        '''
        return [name for name in sorted(self.rddl.domain.action_fluents)]

    @property
    def next_state_fluent_ordering(self) -> List[str]:
        '''The list of next state-fluent names in canonical order.

        Returns:
            List[str]: A list of fluent names.
        '''
        key = lambda x: x.name
        return [cpf.name for cpf in sorted(self.rddl.domain.state_cpfs, key=key)]

    @property
    def interm_fluent_ordering(self) -> List[str]:
        '''The list of intermediate-fluent names in canonical order.

        Returns:
            List[str]: A list of fluent names.
        '''
        interm_fluents = self.rddl.domain.intermediate_fluents.values()
        key = lambda pvar: (pvar.level, pvar.name)
        return [str(pvar) for pvar in sorted(interm_fluents, key=key)]

    @property
    def state_size(self) -> Sequence[Sequence[int]]:
        '''The size of each state fluent in canonical order.

        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        return self._fluent_size(self.initial_state_fluents, self.state_fluent_ordering)

    @property
    def action_size(self) -> Sequence[Sequence[int]]:
        '''The size of each action fluent in canonical order.

        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        return self._fluent_size(self.default_action_fluents, self.action_fluent_ordering)

    @property
    def interm_size(self)-> Sequence[Sequence[int]]:
        '''The size of each intermediate fluent in canonical order.

        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        interm_fluents = self.rddl.domain.intermediate_fluents
        shapes = []
        for name in self.interm_fluent_ordering:
            fluent = interm_fluents[name]
            shape = self._param_types_to_shape(fluent.param_types)
            shapes.append(shape)
        return tuple(shapes)

    @property
    def state_dtype(self) -> Sequence[tf.DType]:
        '''The data type of each state fluent in canonical order.

        Returns:
            Sequence[tf.DType]: A tuple of dtypes representing
            the range of each fluent.
        '''
        return self._fluent_dtype(self.initial_state_fluents, self.state_fluent_ordering)

    @property
    def action_dtype(self) -> Sequence[tf.DType]:
        '''The data type of each action fluent in canonical order.

        Returns:
            Sequence[tf.DType]: A tuple of dtypes representing
            the range of each fluent.
        '''
        return self._fluent_dtype(self.default_action_fluents, self.action_fluent_ordering)

    @property
    def interm_dtype(self) -> Sequence[tf.DType]:
        '''The data type of each intermediate fluent in canonical order.

        Returns:
            Sequence[tf.DType]: A tuple of dtypes representing
            the range of each fluent.
        '''
        interm_fluents = self.rddl.domain.intermediate_fluents
        dtypes = []
        for name in self.interm_fluent_ordering:
            fluent = interm_fluents[name]
            dtype = self._range_type_to_dtype(fluent.range)
            dtypes.append(dtype)
        return tuple(dtypes)

    @property
    def non_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated non-fluents in canonical order.

        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.rddl.domain.non_fluents
        ordering = self.non_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def state_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated state fluents in canonical order.

        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.rddl.domain.state_fluents
        ordering = self.state_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def interm_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated intermediate fluents in canonical order.

        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.rddl.domain.intermediate_fluents
        ordering = self.interm_fluent_ordering
        return self._fluent_params(fluents, ordering)

    @property
    def action_fluent_variables(self) -> FluentParamsList:
        '''Returns the instantiated action fluents in canonical order.

        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        fluents = self.rddl.domain.action_fluents
        ordering = self.action_fluent_ordering
        return self._fluent_params(fluents, ordering)

    def _fluent_params(self, fluents, ordering) -> FluentParamsList:
        '''Returns the instantiated `fluents` for the given `ordering`.

        For each fluent in `fluents`, it instantiates each parameter
        type w.r.t. the contents of the object table.

        Returns:
            Sequence[Tuple[str, List[str]]]: A tuple of pairs of fluent name
            and a list of instantiated fluents represented as strings.
        '''
        variables = []
        for fluent_id in ordering:
            fluent = fluents[fluent_id]
            param_types = fluent.param_types
            objects = ()
            names = []
            if param_types is None:
                names = [fluent.name]
            else:
                objects = tuple(self.object_table[ptype]['objects'] for ptype in param_types)
                for values in itertools.product(*objects):
                    values = ','.join(values)
                    var_name = '{}({})'.format(fluent.name, values)
                    names.append(var_name)
            variables.append((fluent_id, names))
        return tuple(variables)

    @classmethod
    def _fluent_dtype(cls, fluents, ordering) -> Sequence[tf.DType]:
        '''Returns the data types of `fluents` following the given `ordering`.

        Returns:
            Sequence[tf.DType]: A tuple of dtypes representing
            the range of each fluent.
        '''
        dtype = []
        fluents = dict(fluents)
        for name in ordering:
            fluent_dtype = fluents[name].dtype
            dtype.append(fluent_dtype)
        return tuple(dtype)

    @classmethod
    def _fluent_size(cls, fluents, ordering) -> Sequence[Sequence[int]]:
        '''Returns the sizes of `fluents` following the given `ordering`.

        Returns:
            Sequence[Sequence[int]]: A tuple of tuple of integers
            representing the shape and size of each fluent.
        '''
        size = []
        fluents = dict(fluents)
        for name in ordering:
            fluent_shape = fluents[name].shape.fluent_shape
            if fluent_shape == ():
                fluent_shape = (1,)
            size.append(fluent_shape)
        return tuple(size)

    def _build_object_table(self):
        '''Builds the object table for each RDDL type.'''
        types = self.rddl.domain.types
        objects = dict(self.rddl.non_fluents.objects)
        self._object_table = dict()
        for name, value in self.rddl.domain.types:
            if value == 'object':
                objs = objects[name]
                idx = { obj: i for i, obj in enumerate(objs) }
                self._object_table[name] = {
                    'size': len(objs),
                    'idx': idx,
                    'objects': objs
                }

    def _build_preconditions_table(self):
        '''Builds the local action precondition expressions.'''
        self._local_action_preconditions = dict()
        self._global_action_preconditions = []
        action_fluents = self.rddl.domain.action_fluents
        for precond in self.rddl.domain.preconds:
            scope = precond.scope
            action_scope = [action for action in scope if action in action_fluents]
            if len(action_scope) == 1:
                name = action_scope[0]
                self._local_action_preconditions[name] = self._local_action_preconditions.get(name, [])
                self._local_action_preconditions[name].append(precond)
            else:
                self._global_action_preconditions.append(precond)

    def _build_action_bound_constraints_table(self):
        '''Builds the lower and upper action bound constraint expressions.'''
        self._action_lower_bound_constraints = {}
        self._action_upper_bound_constraints = {}

        for name, preconds in self.local_action_preconditions.items():

            for precond in preconds:
                expr_type = precond.etype
                expr_args = precond.args

                if expr_type == ('aggregation', 'forall'):

                    inner_expr = expr_args[1]
                    if inner_expr.etype[0] == 'relational':

                        # lower bound
                        bound = self._extract_lower_bound(name, inner_expr)
                        if bound is not None:
                            self._action_lower_bound_constraints[name] = bound
                            next

                        # upper bound
                        bound = self._extract_upper_bound(name, inner_expr)
                        if bound is not None:
                            self._action_upper_bound_constraints[name] = bound

    def _extract_lower_bound(self, name: str, expr: Expression) -> Optional[Expression]:
        '''Returns the lower bound expression of the action with given `name`.'''
        etype = expr.etype
        args = expr.args
        if etype[1] in ['<=', '<']:
            if args[1].is_pvariable_expression() and args[1].name == name:
                return args[0]
        elif etype[1] in ['>=', '>']:
            if args[0].is_pvariable_expression() and args[0].name == name:
                return args[1]
        return None

    def _extract_upper_bound(self, name: str, expr: Expression) -> Optional[Expression]:
        '''Returns the upper bound expression of the action with given `name`.'''
        etype = expr.etype
        args = expr.args
        if etype[1] in ['<=', '<']:
            if args[0].is_pvariable_expression() and args[0].name == name:
                return args[1]
        elif etype[1] in ['>=', '>']:
            if args[1].is_pvariable_expression() and args[1].name == name:
                return args[0]
        return None

    def _instantiate_pvariables(self,
            pvariables: Dict[str, PVariable],
            ordering: List[str],
            initializer: Optional[InitializerList] = None) -> List[Tuple[str, TensorFluent]]:
        '''Instantiates `pvariables` given an initialization list and
        returns a list of TensorFluents in the given `ordering`.

        Returns:
            List[Tuple[str, TensorFluent]]: A list of pairs of fluent name and fluent tensor.
        '''
        if initializer is not None:
            init = dict()
            for ((name, args), value) in initializer:
                arity = len(args) if args is not None else 0
                name = '{}/{}'.format(name, arity)
                init[name] = init.get(name, [])
                init[name].append((args, value))

        fluents = []

        for name in ordering:
            pvar = pvariables[name]
            shape = self._param_types_to_shape(pvar.param_types)
            dtype = self._range_type_to_dtype(pvar.range)
            fluent = np.full(shape, pvar.default)

            if initializer is not None:
                for args, val in init.get(name, []):
                    if args is not None:
                        idx = []
                        for ptype, arg in zip(pvar.param_types, args):
                            idx.append(self.object_table[ptype]['idx'][arg])
                        idx = tuple(idx)
                        fluent[idx] = val
                    else:
                        fluent = val

            with self.graph.as_default():
                t = tf.constant(fluent, dtype=dtype, name=name)
                scope = [None] * len(t.shape)
                fluent = TensorFluent(t, scope, batch=False)
                fluent_pair = (name, fluent)
                fluents.append(fluent_pair)

        return fluents

    def _instantiate_non_fluents(self):
        '''Returns the non-fluents instantiated.'''
        non_fluents = self.rddl.domain.non_fluents
        initializer = self.rddl.non_fluents.init_non_fluent
        self._non_fluents = self._instantiate_pvariables(non_fluents, self.non_fluent_ordering, initializer)
        return self._non_fluents

    def _instantiate_initial_state_fluents(self):
        '''Returns the initial state-fluents instantiated.'''
        state_fluents = self.rddl.domain.state_fluents
        initializer = self.rddl.instance.init_state
        self._initial_state_fluents = self._instantiate_pvariables(state_fluents, self.state_fluent_ordering, initializer)
        return self._initial_state_fluents

    def _instantiate_default_action_fluents(self):
        '''Returns the default action-fluents instantiated.'''
        action_fluents = self.rddl.domain.action_fluents
        self._default_action_fluents = self._instantiate_pvariables(action_fluents, self.action_fluent_ordering)
        return self._default_action_fluents

    def _compile_batch_fluents(self,
            fluents: List[Tuple[str, TensorFluent]],
            batch_size: int) -> Sequence[tf.Tensor]:
        '''Compiles `fluents` into tensors with given `batch_size`.

        Returns:
            Sequence[tf.Tensor]: A tuple of tensors with first dimensio
            corresping to the batch size.
        '''
        batch_fluents = []
        with self.graph.as_default():
            for name, fluent in fluents:
                t = tf.stack([fluent.tensor] * batch_size, name=name)
                if t.shape.ndims == 1:
                    t = tf.expand_dims(t, -1)
                batch_fluents.append(t)
        return tuple(batch_fluents)

    def _compile_expression(self,
            expr: Expression,
            scope: Dict[str, TensorFluent],
            batch_size: Optional[int] = None) -> TensorFluent:
        '''Compile the expression `expr` into a TensorFluent
        in the given `scope` with optional batch size.

        Args:
            expr (:obj:`tfrddlsim.expr.Expression`): A RDDL expression.
            scope (Dict[str, :obj:`tfrddlsim.fluent.TensorFluent`]): A fluent scope.
            batch_size (Optional[size]): The batch size.

        Returns:
            :obj:`tfrddlsim.fluent.TensorFluent`: A TensorFluent.
        '''
        etype = expr.etype
        args = expr.args

        with self.graph.as_default():

            if etype[0] == 'number':
                return TensorFluent.constant(args)
            elif etype[0] == 'pvar':
                name = expr._pvar_to_name(args)
                if name not in scope:
                    raise ValueError('Variable {} not in scope.'.format(name))
                fluent = scope[name]
                scope = args[1] if args[1] is not None else []
                if isinstance(fluent, TensorFluent):
                    fluent = TensorFluent(fluent.tensor, scope, batch=fluent.batch)
                elif isinstance(fluent, tf.Tensor):
                    fluent = TensorFluent(fluent, scope, batch=self.batch_mode)
                else:
                    raise ValueError('Variable in scope must be TensorFluent-like: {}'.format(fluent))
                return fluent
            elif etype[0] == 'randomvar':
                if etype[1] == 'Normal':
                    mean = self._compile_expression(args[0], scope)
                    variance = self._compile_expression(args[1], scope)
                    return TensorFluent.Normal(mean, variance, batch_size)
                elif etype[1] == 'Uniform':
                    low = self._compile_expression(args[0], scope)
                    high = self._compile_expression(args[1], scope)
                    return TensorFluent.Uniform(low, high, batch_size)
                elif etype[1] == 'Exponential':
                    mean = self._compile_expression(args[0], scope)
                    return TensorFluent.Exponential(mean, batch_size)
                elif etype[1] == 'Gamma':
                    shape = self._compile_expression(args[0], scope)
                    scale = self._compile_expression(args[1], scope)
                    return TensorFluent.Gamma(shape, scale, batch_size)
            elif etype[0] == 'arithmetic':
                if etype[1] == '+':
                    if len(args) == 1:
                        op1 = self._compile_expression(args[0], scope)
                        return op1
                    else:
                        op1 = self._compile_expression(args[0], scope)
                        op2 = self._compile_expression(args[1], scope)
                        return op1 + op2
                elif etype[1] == '-':
                    if len(args) == 1:
                        op1 = self._compile_expression(args[0], scope)
                        return -op1
                    else:
                        op1 = self._compile_expression(args[0], scope)
                        op2 = self._compile_expression(args[1], scope)
                        return op1 - op2
                elif etype[1] == '*':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 * op2
                elif etype[1] == '/':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 / op2
            elif etype[0] == 'boolean':
                if etype[1] in ['^', '&']:
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 & op2
                elif etype[1] == '|':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 | op2
                elif etype[1] == '=>':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return ~op1 | op2
                elif etype[1] == '<=>':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return (op1 & op2) | (~op1 & ~op2)
                elif etype[1] == '~':
                    op = self._compile_expression(args[0], scope)
                    return ~op
            elif etype[0] == 'relational':
                if etype[1] == '<=':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 <= op2
                elif etype[1] == '<':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 < op2
                elif etype[1] == '>=':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 >= op2
                elif etype[1] == '>':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 > op2
                elif etype[1] == '==':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 == op2
                elif etype[1] == '~=':
                    op1 = self._compile_expression(args[0], scope)
                    op2 = self._compile_expression(args[1], scope)
                    return op1 != op2
            elif etype[0] == 'func':
                if etype[1] == 'abs':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.abs(x)
                elif etype[1] == 'exp':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.exp(x)
                elif etype[1] == 'log':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.log(x)
                elif etype[1] == 'sqrt':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.sqrt(x)
                elif etype[1] == 'cos':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.cos(x)
                elif etype[1] == 'sin':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.sin(x)
                elif etype[1] == 'round':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.round(x)
                elif etype[1] == 'ceil':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.ceil(x)
                elif etype[1] == 'floor':
                    x = self._compile_expression(args[0], scope)
                    return TensorFluent.floor(x)
                elif etype[1] == 'pow':
                    x = self._compile_expression(args[0], scope)
                    y = self._compile_expression(args[1], scope)
                    return TensorFluent.pow(x, y)
                elif etype[1] == 'max':
                    x = self._compile_expression(args[0], scope)
                    y = self._compile_expression(args[1], scope)
                    return TensorFluent.maximum(x, y)
                elif etype[1] == 'min':
                    x = self._compile_expression(args[0], scope)
                    y = self._compile_expression(args[1], scope)
                    return TensorFluent.minimum(x, y)
            elif etype[0] == 'control':
                if etype[1] == 'if':
                    condition = self._compile_expression(args[0], scope)
                    true_case = self._compile_expression(args[1], scope)
                    false_case = self._compile_expression(args[2], scope)
                    return TensorFluent.if_then_else(condition, true_case, false_case)
            elif etype[0] == 'aggregation':
                if etype[1] not in ['sum', 'prod', 'exists', 'forall']:
                    raise ValueError('Unkown aggregation function {}.'.format(etype[1]))
                typed_var_list = args[:-1]
                vars_list = [var for _, (var, _) in typed_var_list]
                expr = args[-1]
                x = self._compile_expression(expr, scope)
                if etype[1] == 'sum':
                    return x.sum(vars_list=vars_list)
                elif etype[1] == 'prod':
                    return x.prod(vars_list=vars_list)
                elif etype[1] == 'exists':
                    return x.exists(vars_list=vars_list)
                elif etype[1] == 'forall':
                    return x.forall(vars_list=vars_list)

    @classmethod
    def _range_type_to_dtype(cls, range_type: str) -> Optional[tf.DType]:
        '''Maps RDDL range types to TensorFlow dtypes.'''
        dtype = None
        if range_type == 'real':
            dtype = tf.float32
        elif range_type == 'int':
            dtype = tf.int32
        elif range_type == 'bool':
            dtype = tf.bool
        return dtype

    def _param_types_to_shape(self, param_types: Optional[str]) -> Sequence[int]:
        '''Returns the fluent shape given its `param_types`.'''
        param_types = [] if param_types is None else param_types
        shape = tuple(self.object_table[ptype]['size'] for ptype in param_types)
        return shape



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.

from pyrddl.parser import RDDLParser

from tfrddlsim.rddl2tf.compiler import Compiler
from tfrddlsim.simulation.transition_simulator import ActionSimulationCell

import tensorflow as tf
import unittest


class TestActionSimulationCell(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        parser = RDDLParser()
        parser.build()

        with open('rddl/Reservoir.rddl', mode='r') as file:
            RESERVOIR = file.read()
            cls.rddl1 = parser.parse(RESERVOIR)

        with open('rddl/Navigation.rddl', mode='r') as file:
            NAVIGATION = file.read()
            cls.rddl2 = parser.parse(NAVIGATION)

        cls.batch_size = 10

    def setUp(self):
        self.compiler1 = Compiler(self.rddl1, batch_mode=True)
        self.cell1 = ActionSimulationCell(self.compiler1)
        self.initial_state1 = self.compiler1.compile_initial_state(batch_size=self.batch_size)
        self.default_action1 = self.compiler1.compile_default_action(batch_size=1)

        self.compiler2 = Compiler(self.rddl2, batch_mode=True)
        self.cell2 = ActionSimulationCell(self.compiler2)
        self.initial_state2 = self.compiler2.compile_initial_state(batch_size=self.batch_size)
        self.default_action2 = self.compiler2.compile_default_action(batch_size=1)

    def test_state_size(self):
        state_size1 = self.cell1.state_size
        self.assertIsInstance(state_size1, tuple)
        self.assertEqual(len(state_size1), len(self.initial_state1))
        for shape, tensor in zip(state_size1, self.initial_state1):
            self.assertListEqual(list(shape), tensor.shape.as_list()[1:])

        state_size2 = self.cell2.state_size
        self.assertIsInstance(state_size2, tuple)
        self.assertEqual(len(state_size2), len(self.initial_state2))
        for shape, tensor in zip(state_size2, self.initial_state2):
            self.assertListEqual(list(shape), tensor.shape.as_list()[1:])

    def test_next_state(self):
        (_, next_state1) = self.cell1(self.default_action1, self.initial_state1)
        self.assertIsInstance(next_state1, tuple)
        self.assertEqual(len(next_state1), len(self.initial_state1))
        for t1, t2 in zip(self.initial_state1, next_state1):
            self.assertIsInstance(t2, tf.Tensor)
            self.assertEqual(t1.shape, t2.shape)
            self.assertEqual(t1.dtype, t2.dtype)

        (_, next_state2) = self.cell2(self.default_action2, self.initial_state2)
        self.assertIsInstance(next_state2, tuple)
        self.assertEqual(len(next_state2), len(self.initial_state2))
        for t1, t2 in zip(self.initial_state2, next_state2):
            self.assertIsInstance(t2, tf.Tensor)
            self.assertEqual(t1.shape, t2.shape)
            self.assertEqual(t1.dtype, t2.dtype)

    def test_output_size(self):
        output_size1 = self.cell1.output_size
        output_size2 = self.cell2.output_size
        self.assertTupleEqual(output_size1, (1,))
        self.assertTupleEqual(output_size2, (1,))

    def test_output(self):
        (reward1, _) = self.cell1(self.default_action1, self.initial_state1)
        self.assertIsInstance(reward1, tuple)
        self.assertEqual(len(reward1), 1)
        self.assertIsInstance(reward1[0], tf.Tensor)
        self.assertListEqual(reward1[0].shape.as_list(), [self.batch_size, 1])
        self.assertEqual(reward1[0].dtype, tf.float32)

        (reward2, _) = self.cell2(self.default_action2, self.initial_state2)
        self.assertIsInstance(reward2, tuple)
        self.assertEqual(len(reward2), 1)
        self.assertIsInstance(reward2[0], tf.Tensor)
        self.assertListEqual(reward2[0].shape.as_list(), [self.batch_size, 1])
        self.assertEqual(reward2[0].dtype, tf.float32)



# This file is part of tf-rddlsim.

# tf-rddlsim is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# tf-rddlsim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with tf-rddlsim. If not, see <http://www.gnu.org/licenses/>.


from tfrddlsim.rddl2tf.fluentscope import TensorFluentScope
from tfrddlsim.rddl2tf.fluentshape import TensorFluentShape

import tensorflow as tf

from typing import Callable, List, Optional, Sequence, Union

Value = Union[bool, int, float]


class TensorFluent(object):
    '''TensorFluent is a wrapper for ``tf.Tensor`` objects.

    Each RDDL fluent is compiled to a ``TensorFluent`` after instantiation.
    A ``TensorFluent`` stores metadata needed for compiling RDDL expressions
    into TensorFlow's graph ops and for evaluating fluents in batch mode.

    Args:
        tensor: A tensor op representing the fluent in the graph.
        scope: The fluent's argument scope in an expression.
        batch: Batch mode flag.

    Attributes:
        tensor (:obj:`tf.Tensor`): A tensor op representing the fluent in the graph.
        scope (:obj:`tfrddlsim.fluentscope.TensorFluentScope`): The argument scope of the fluent in the expression.
        shape (:obj:`tfrddlsim.fluentshape.TensorFluentShape`): The fluent shape and dimensions.
    '''

    def __init__(self, tensor: tf.Tensor, scope: List[str], batch: bool = False) -> None:
        self.tensor = tensor
        self.scope = TensorFluentScope(scope)
        self.shape = TensorFluentShape(tensor.shape, batch)

    @property
    def batch(self) -> bool:
        '''Returns True if in batch mode. False, otherwise.'''
        return self.shape.batch

    @property
    def dtype(self) -> tf.DType:
        '''Returns the fluent's data type.'''
        return self.tensor.dtype

    @property
    def name(self) -> str:
        '''Returns the tensor's name.'''
        return self.tensor.name

    @classmethod
    def constant(cls,
            value: Value,
            dtype: tf.DType = tf.float32) -> 'TensorFluent':
        '''Returns a constant `value` TensorFluent with given `dtype`.

        Args:
            value: The constant value.
            dtype: The output's data type.

        Returns:
            A constant TensorFluent.
        '''
        t = tf.constant(value, dtype=dtype)
        scope = [] # type: List
        batch = False
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def Normal(cls,
            mean: 'TensorFluent', variance: 'TensorFluent',
            batch_size: Optional[int] = None) -> 'TensorFluent':
        '''Returns a TensorFluent for the Normal sampling op with given mean and variance.

        Args:
            mean: The mean parameter of the Normal distribution.
            variance: The variance parameter of the Normal distribution.
            batch_size: The size of the batch (optional).

        Returns:
            A TensorFluent sample drawn from the Normal distribution.

        Raises:
            ValueError: If parameters do not have the same scope.
        '''
        if mean.scope != variance.scope:
            raise ValueError('Normal distribution: parameters must have same scope!')
        loc = mean.tensor
        scale = tf.sqrt(variance.tensor)
        dist = tf.distributions.Normal(loc, scale)
        batch = mean.batch or variance.batch
        if not batch and batch_size is not None:
            t = dist.sample(batch_size)
            batch = True
        else:
            t = dist.sample()
        scope = mean.scope.as_list()
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def Uniform(cls,
            low: 'TensorFluent', high: 'TensorFluent',
            batch_size: Optional[int] = None) -> 'TensorFluent':
        '''Returns a TensorFluent for the Uniform sampling op with given low and high parameters.

        Args:
            low: The low parameter of the Uniform distribution.
            high: The high parameter of the Uniform distribution.
            batch_size: The size of the batch (optional).

        Returns:
            A TensorFluent sample drawn from the Uniform distribution.

        Raises:
            ValueError: If parameters do not have the same scope.
        '''
        if low.scope != high.scope:
            raise ValueError('Uniform distribution: parameters must have same scope!')
        low = low.tensor
        high = high.tensor
        dist = tf.distributions.Uniform(low, high)
        batch = low.batch or high.batch
        if not batch and batch_size is not None:
            t = dist.sample(batch_size)
            batch = True
        else:
            t = dist.sample()
        scope = low.scope.as_list()
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def Exponential(cls,
            mean: 'TensorFluent',
            batch_size: Optional[int] = None) -> 'TensorFluent':
        '''Returns a TensorFluent for the Exponential sampling op with given mean parameter.

        Args:
            mean: The mean parameter of the Exponential distribution.
            batch_size: The size of the batch (optional).

        Returns:
            A TensorFluent sample drawn from the Exponential distribution.
        '''
        rate = 1 / mean.tensor
        dist = tf.distributions.Exponential(rate)
        batch = mean.batch
        if not batch and batch_size is not None:
            t = dist.sample(batch_size)
            batch = True
        else:
            t = dist.sample()
        scope = mean.scope.as_list()
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def Gamma(cls,
            shape: 'TensorFluent',
            scale: 'TensorFluent',
            batch_size: Optional[int] = None) -> 'TensorFluent':
        '''Returns a TensorFluent for the Gamma sampling op with given shape and scale parameters.

        Args:
            shape: The shape parameter of the Gamma distribution.
            scale: The scale parameter of the Gamma distribution.
            batch_size: The size of the batch (optional).

        Returns:
            A TensorFluent sample drawn from the Uniform distribution.

        Raises:
            ValueError: If parameters do not have the same scope.
        '''
        if shape.scope != scale.scope:
            raise ValueError('Gamma distribution: parameters must have same scope!')
        concentration = shape.tensor
        rate = 1 / scale.tensor
        dist = tf.distributions.Gamma(concentration, rate)
        batch = shape.batch or scale.batch
        if not batch and batch_size is not None:
            t = dist.sample(batch_size)
            batch = True
        else:
            t = dist.sample()
        scope = shape.scope.as_list()
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def abs(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the abs function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the abs function.
        '''
        return cls._unary_op(x, tf.abs, tf.float32)

    @classmethod
    def exp(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the exp function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the exp function.
        '''
        return cls._unary_op(x, tf.exp, tf.float32)

    @classmethod
    def log(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the log function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the log function.
        '''
        return cls._unary_op(x, tf.log, tf.float32)

    @classmethod
    def sqrt(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the sqrt function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the sqrt function.
        '''
        return cls._unary_op(x, tf.sqrt, tf.float32)

    @classmethod
    def cos(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the cos function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the cos function.
        '''
        return cls._unary_op(x, tf.cos, tf.float32)

    @classmethod
    def sin(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the sin function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the sin function.
        '''
        return cls._unary_op(x, tf.sin, tf.float32)

    @classmethod
    def tan(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the tan function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the tan function.
        '''
        return cls._unary_op(x, tf.tan, tf.float32)

    @classmethod
    def round(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the round function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the round function.
        '''
        return cls._unary_op(x, tf.round, tf.float32)

    @classmethod
    def ceil(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the ceil function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the ceil function.
        '''
        return cls._unary_op(x, tf.ceil, tf.float32)

    @classmethod
    def floor(cls, x: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the floor function.

        Args:
            x: The input fluent.

        Returns:
            A TensorFluent wrapping the floor function.
        '''
        return cls._unary_op(x, tf.floor, tf.float32)

    @classmethod
    def pow(cls, x: 'TensorFluent', y: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the pow function.TensorFluent

        Args:
            x: The first operand.
            y: The second operand.

        Returns:
            A TensorFluent wrapping the pow function.
        '''
        return cls._binary_op(x, y, tf.pow, tf.float32)

    @classmethod
    def maximum(cls, x: 'TensorFluent', y: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the maximum function.TensorFluent

        Args:
            x: The first operand.
            y: The second operand.

        Returns:
            A TensorFluent wrapping the maximum function.
        '''
        return cls._binary_op(x, y, tf.maximum, tf.float32)

    @classmethod
    def minimum(cls, x: 'TensorFluent', y: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the minimum function.

        Args:
            x: The first operand.
            y: The second operand.

        Returns:
            A TensorFluent wrapping the minimum function.
        '''
        return cls._binary_op(x, y, tf.minimum, tf.float32)

    @classmethod
    def if_then_else(cls,
            condition: 'TensorFluent',
            true_case: 'TensorFluent',
            false_case: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the control op if-then-else.

        Args:
            condition: Boolean fluent for the if condition.
            true_case: Fluent returned in the true clause.
            false_case: Fluent returned in the false clause.

        Returns:
            A TensorFluent wrapping the if-then-else control statement.

        Raises:
            ValueError: If cases don't have same shape.
        '''
        condition_tensor = condition.tensor
        true_case_tensor = true_case.tensor
        false_case_tensor = false_case.tensor

        if true_case.shape != false_case.shape:
            if true_case.shape.as_list() == []:
                true_case_tensor = tf.fill(false_case.shape.as_list(), true_case.tensor)
            elif false_case.shape.as_list() == []:
                false_case_tensor = tf.fill(true_case.shape.as_list(), false_case.tensor)

        if true_case_tensor.shape != false_case_tensor.shape:
            raise ValueError('TensorFluent.if_then_else: cases must be of same shape!')

        t = tf.where(condition_tensor, x=true_case_tensor, y=false_case_tensor)
        scope = condition.scope.as_list()

        batch = condition.batch
        # if (not batch) and (condition.batch or true_case.batch or false_case.batch):
        #     raise ValueError('TensorFluent.if_then_else: cases must be batch compatible!')

        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def _binary_op(cls,
            x: 'TensorFluent',
            y: 'TensorFluent',
            op: Callable[[tf.Tensor, tf.Tensor], tf.Tensor],
            dtype: tf.DType) -> 'TensorFluent':
        '''Returns a TensorFluent for the binary `op` applied to fluents `x` and `y`.

        Args:
            x: The first operand.
            y: The second operand.
            op: The binary operator.
            dtype: The output's data type.

        Returns:
            A TensorFluent wrapping the binary operator's output.
        '''
        # scope
        s1 = x.scope.as_list()
        s2 = y.scope.as_list()
        scope, perm1, perm2 = TensorFluentScope.broadcast(s1, s2)
        if x.batch and perm1 != []:
            perm1 = [0] + [p+1 for p in perm1]
        if y.batch and perm2 != []:
            perm2 = [0] + [p+1 for p in perm2]
        x = x.transpose(perm1)
        y = y.transpose(perm2)

        # shape
        reshape1, reshape2 = TensorFluentShape.broadcast(x.shape, y.shape)
        if reshape1 is not None:
            x = x.reshape(reshape1)
        if reshape2 is not None:
            y = y.reshape(reshape2)

        # dtype
        x = x.cast(dtype)
        y = y.cast(dtype)

        # operation
        t = op(x.tensor, y.tensor)

        # batch
        batch = x.batch or y.batch

        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def _unary_op(cls,
            x: 'TensorFluent',
            op: Callable[[tf.Tensor], tf.Tensor],
            dtype: tf.DType) -> 'TensorFluent':
        '''Returns a TensorFluent for the unary `op` applied to fluent `x`.

        Args:
            x: The input fluent.
            op: The unary operation.
            dtype: The output's data type.

        Returns:
            A TensorFluent wrapping the unary operator's output.
        '''
        x = x.cast(dtype)
        t = op(x.tensor)
        scope = x.scope.as_list()
        batch = x.batch
        return TensorFluent(t, scope, batch=batch)

    @classmethod
    def _aggregation_op(cls,
            op: Callable[[tf.Tensor, Optional[Sequence[int]]], tf.Tensor],
            x: 'TensorFluent',
            vars_list: List[str]) -> 'TensorFluent':
        '''Returns a TensorFluent for the aggregation `op` applied to fluent `x`.

        Args:
            op: The aggregation operation.
            x: The input fluent.
            vars_list: The list of variables to be aggregated over.

        Returns:
            A TensorFluent wrapping the aggregation operator's output.
        '''
        axis = []
        for var in vars_list:
            if var in x.scope.as_list():
                ax = x.scope.index(var)
                if x.batch:
                    ax += 1
                axis.append(ax)
        t = op(x.tensor, axis)

        scope = []
        for var in x.scope.as_list():
            if var not in vars_list:
                scope.append(var)

        batch = x.batch

        return TensorFluent(t, scope, batch=batch)

    def cast(self, dtype: tf.DType) -> 'TensorFluent':
        '''Returns a TensorFluent for the cast operation with given `dtype`.

        Args:
            dtype: The output's data type.

        Returns:
            A TensorFluent wrapping the cast operation.
        '''
        t = self.tensor if self.tensor.dtype == dtype else tf.cast(self.tensor, dtype)
        scope = self.scope.as_list()
        batch = self.batch
        return TensorFluent(t, scope, batch=batch)

    def reshape(self, shape: tf.TensorShape) -> 'TensorFluent':
        '''Returns a TensorFluent for the reshape operation with given `shape`.

        Args:
            shape: The output's shape.

        Returns:
            A TensorFluent wrapping the reshape operation.
        '''
        t = tf.reshape(self.tensor, shape)
        scope = self.scope.as_list()
        batch = self.batch
        return TensorFluent(t, scope, batch=batch)

    def transpose(self, permutation: Optional[List[int]] = None) -> 'TensorFluent':
        '''Returns a TensorFluent for the transpose operation with given `permutation`.

        Args:
            permutation: The output's shape permutation.

        Returns:
            A TensorFluent wrapping the transpose operation.
        '''
        t = tf.transpose(self.tensor, permutation) if permutation != [] else self.tensor
        scope = self.scope.as_list()
        batch = self.batch
        return TensorFluent(t, scope, batch=batch)

    def sum(self, vars_list: List[str]) -> 'TensorFluent':
        '''Returns the TensorFluent for the sum aggregation function.

        Args:
            vars_list: The list of variables to be aggregated over.

        Returns:
            A TensorFluent wrapping the sum aggregation function.
        '''
        return self._aggregation_op(tf.reduce_sum, self, vars_list)

    def prod(self, vars_list: List[str]) -> 'TensorFluent':
        '''Returns the TensorFluent for the prod aggregation function.

        Args:
            vars_list: The list of variables to be aggregated over.

        Returns:
            A TensorFluent wrapping the prod aggregation function.
        '''
        return self._aggregation_op(tf.reduce_prod, self, vars_list)

    def forall(self, vars_list: List[str]) -> 'TensorFluent':
        '''Returns the TensorFluent for the forall aggregation function.

        Args:
            vars_list: The list of variables to be aggregated over.

        Returns:
            A TensorFluent wrapping the forall aggregation function.
        '''
        return self._aggregation_op(tf.reduce_all, self, vars_list)

    def exists(self, vars_list: List[str]) -> 'TensorFluent':
        '''Returns the TensorFluent for the exists aggregation function.

        Args:
            vars_list: The list of variables to be aggregated over.

        Returns:
            A TensorFluent wrapping the exists aggregation function.
        '''
        return self._aggregation_op(tf.reduce_any, self, vars_list)

    def __neg__(self):
        '''Returns a TensorFluent for the unary negative operator.

        Args:
            self: The operand

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._unary_op(self, tf.negative, tf.float32)

    def __add__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the addition arithmetic operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.add, tf.float32)

    def __sub__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the subtraction arithmetic operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.subtract, tf.float32)

    def __mul__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the multiplication arithmetic operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.multiply, tf.float32)

    def __truediv__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the division arithmetic operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.divide, tf.float32)

    def __and__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the and logical operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.logical_and, tf.bool)

    def __or__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the or logical operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.logical_or, tf.bool)

    def __xor__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the xor logical operator.

        Args:
            self: The first operand.
            other: The second operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._binary_op(self, other, tf.logical_xor, tf.bool)

    def __invert__(self) -> 'TensorFluent':
        '''Returns a TensorFluent for the not logical operator.

        Args:
            self: The operand.

        Returns:
            A TensorFluent wrapping the operator's output.
        '''
        return self._unary_op(self, tf.logical_not, tf.bool)

    def __le__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the less-than-or-equal relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.less_equal, tf.float32)

    def __lt__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the less-then relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.less, tf.float32)

    def __ge__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the greater-then-or-equal relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.greater_equal, tf.float32)

    def __gt__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the greater-than relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.greater, tf.float32)

    def __eq__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the equal relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.equal, tf.float32)

    def __ne__(self, other: 'TensorFluent') -> 'TensorFluent':
        '''Returns a TensorFluent for the not-equal relational operator.

        Args:
            self: The first operand.
            other: The second operand.
        '''
        return self._binary_op(self, other, tf.not_equal, tf.float32)

    def __str__(self) -> str:
        '''Returns TensorFluent's string representation.'''
        return 'TensorFluent("{}", dtype={}, {}, {})'.format(self.name, repr(self.dtype), self.scope, self.shape)
